﻿#include <nn.h>
#include <nn/fnd.h>
#include <nn/cfg.h>
#include <nn/jpeg.h>

#include "ImageIO.h"


namespace {


// メインヒープサイズ
const std::size_t   MAIN_MEMORY_SIZE = 2*1024*1024;

// 取り扱う画像
const int IMAGE_WIDTH  = 640;
const int IMAGE_HEIGHT = 480;
const nn::jpeg::CTR::PixelFormat IMAGE_FORMAT = nn::jpeg::CTR::PIXEL_FORMAT_CTR_RGB565;

// タイトルごとに割り当てられる 20bit の ユニークID を指定してください。
const u32 TITLE_UNIQUE_ID = 0xFF000;

/*---------------------------------------------------------------------------*
  アロケータ
 *---------------------------------------------------------------------------*/
class MyAllocator : public nn::fnd::IAllocator
{
public:
    void    Initialize(){ m_Heap.Initialize(nn::os::GetHeapAddress(), nn::os::GetHeapSize()); }
    void    Finalize(){ m_Heap.Finalize(); }

    virtual void* Allocate(size_t size, s32 alignment){ return m_Heap.Allocate(size, alignment); }
    virtual void Free(void* memory){ m_Heap.Free(memory); }

private:
    nn::fnd::ExpHeap    m_Heap;
};

MyAllocator s_Allocator;

/*---------------------------------------------------------------------------*
  JPEGの保存と読み込みの処理
 *---------------------------------------------------------------------------*/
void TestSaveLoad()
{
    imgdb::JpegSaver    jpgSaver;
    imgdb::JpegLoader   jpgLoader;

    // 適当な画像を作成してJPEGで保存する
    {
        // 適当な画像を作成する
        void* pImageData = 0;
        const std::size_t nImageSize = imgdb::JpegLoader::ComputeImageSize(
                                            IMAGE_WIDTH, IMAGE_HEIGHT, IMAGE_FORMAT);
        pImageData = s_Allocator.Allocate(nImageSize, 32);
        std::memset(pImageData, 0xff, nImageSize); // 真っ白の画像

        // JPEG保存の各種設定を行う
        {
            //---------- 必須設定 ----------//
            // 保存する画像
            jpgSaver.SetRaw(pImageData, IMAGE_WIDTH, IMAGE_HEIGHT, IMAGE_FORMAT);
            // タイトルユニークID
            jpgSaver.SetTitleUniqueId(TITLE_UNIQUE_ID);

            //---------- 任意設定 ----------//
            // ユーザーメーカーノート
            const char* pUserMakerNoteData = "This is UserMakerNote!";
            jpgSaver.SetUserMakerNote(pUserMakerNoteData, strlen(pUserMakerNoteData)+1);
        }

        // 保存
        {
            const imgdb::Result eResult = jpgSaver.Save();
            if(imgdb::ResultSuccess == eResult){
                NN_LOG("Saved! %ls\n", jpgSaver.GetSavedFilePath());
            }
            else if(imgdb::ResultErrorNotExistStorage == eResult ||
                    imgdb::ResultErrorWriteProtect == eResult ||
                    imgdb::ResultErrorInsufficientStorageSpace == eResult ||
                    imgdb::ResultErrorFileOperationFailure == eResult)
            {
                // ハンドリングすべきエラー
                NN_PANIC("Error! Result=%d\n", eResult);
            }
            else{
                NN_PANIC("Error! Result=%d\n", eResult);
            }
        }

        s_Allocator.Free(pImageData);
    }

    // 先ほど保存した画像を読み込む
    {
        const std::size_t nUserMakerNoteBufSize = 32;
        char  nUserMakerNoteBuf[nUserMakerNoteBufSize];

        // 画像を読み込むバッファを作成する
        void* pImageData = 0;
        const std::size_t nImageSize = imgdb::JpegLoader::ComputeImageSize(
                                            IMAGE_WIDTH, IMAGE_HEIGHT, IMAGE_FORMAT);
        pImageData = s_Allocator.Allocate(nImageSize, 32);
        std::memset(pImageData, 0x00, nImageSize);

        // JPEG読み込みの各種設定を行う
        {
            //---------- 必須設定 ----------//
            jpgLoader.SetFilePath(jpgSaver.GetSavedFilePath());
            jpgLoader.SetBuffer(pImageData, nImageSize);
            jpgLoader.SetFormat(IMAGE_FORMAT);

            //---------- 任意設定 ----------//
            jpgLoader.SetUserMakerNoteBuffer(nUserMakerNoteBuf, nUserMakerNoteBufSize);
        }

        // 読み込み
        {
            const imgdb::Result eResult = jpgLoader.Load();
            if(imgdb::ResultSuccess == eResult){
                NN_LOG("Loaded!\n");
            }
            else if(imgdb::ResultErrorNotExistStorage == eResult ||
                    imgdb::ResultErrorWriteProtect == eResult ||
                    imgdb::ResultErrorInsufficientStorageSpace == eResult ||
                    imgdb::ResultErrorFileOperationFailure == eResult)
            {
                // ハンドリングすべきエラー
                NN_PANIC("Error! Result=%d\n", eResult);
            }
            else{
                NN_PANIC("Error! Result=%d\n", eResult);
            }
        }

        // 設定されている情報を取り出す
        {
            NN_LOG(" TitleUniqueId              : 0x%08x\n", jpgLoader.GetTitleUniqueId());

            NN_LOG(" ScreenshotFlag             : %s\n", jpgLoader.GetScreenshotFlag() ? "true" : "false");
            NN_LOG(" AdjustablePictureFlag      : %s\n", jpgLoader.GetAdjustablePictureFlag() ? "true" : "false");

            if(jpgLoader.IsValidShotBodyId()){
                NN_LOG(" ShotBodyId                 : 0x%08x\n", jpgLoader.GetShotBodyId());
            }
            else{
                NN_LOG(" ShotBodyId                 : Invalid\n");
            }
            if(jpgLoader.IsValidEditorBodyId()){
                NN_LOG(" EditorBodyId               : 0x%08x\n", jpgLoader.GetEditorBodyId());
            }
            else{
                NN_LOG(" EditorBodyId               : Invalid\n");
            }

            NN_LOG(" UserMakerNote              : %dByte\n", jpgLoader.GetUserMakerNoteDataSize());
            if(0 < jpgLoader.GetUserMakerNoteDataSize()){
                nUserMakerNoteBuf[nUserMakerNoteBufSize-1] = '\0';
                NN_LOG( " \"%s\"\n", nUserMakerNoteBuf);
            }
        }

        s_Allocator.Free(pImageData);
    }

    // SDカードの写真を列挙する
    {
        NN_LOG("\n");
        imgdb::search::SearchPicture();
        for(int i = 0; i < imgdb::search::GetResultNum(); ++i){
            static const char* pImgKindStr[imgdb::imgio::ImageKindNum+1] = { "JPG", "MPO", "???", };

            // タイトルユニークIDを取得する
            const u32 nId = imgdb::search::GetTitleUniqueId(i);

            // パスを取得する
            wchar_t cPath[imgdb::imgio::c_nMaxPathSize];
            imgdb::search::GetPath(cPath, sizeof(cPath)/sizeof(cPath[0]), i);

            // 写真の種類を取得する
            const imgdb::imgio::ImageKind eKind = imgdb::search::GetImageKind(i);

            // 撮影日時を取得する
            const nn::fnd::DateTimeParameters dtp = imgdb::search::GetDateTime(i).GetParameters();

            NN_LOG(
                "0x%05x \"%ls\" [%s] %04d/%02d/%02d %02d:%02d.%02d\n",
                nId, cPath, pImgKindStr[eKind], dtp.year, dtp.month, dtp.day, dtp.hour, dtp.minute, dtp.second);
        }
        NN_LOG("\n");
    }

    // 保存した写真を削除する
    {
        const imgdb::Result eResult =
                imgdb::util::DeletePicture(jpgSaver.GetSavedFilePath(), TITLE_UNIQUE_ID);
        if(imgdb::ResultSuccess == eResult){
            NN_LOG("Deleted! %ls\n", jpgSaver.GetSavedFilePath());
        }
        else if(imgdb::ResultErrorNotExistStorage == eResult ||
                imgdb::ResultErrorWriteProtect == eResult ||
                imgdb::ResultErrorInsufficientStorageSpace == eResult ||
                imgdb::ResultErrorFileOperationFailure == eResult)
        {
            // ハンドリングすべきエラー
            NN_PANIC("Error! Result=%d\n", eResult);
        }
        else{
            NN_PANIC("Error! Result=%d\n", eResult);
        }
    }
}

} // namespace

//----------------------------------------------------------------
extern "C" void nninitStartUp()
{
}

/*---------------------------------------------------------------------------*
  メイン処理
 *---------------------------------------------------------------------------*/
extern "C" void nnMain()
{
	nn::cfg::Initialize(); // ImageDb内部でnn::cfgの機能を利用しているため要初期化
	nn::fs::Initialize();

    nn::os::SetHeapSize(MAIN_MEMORY_SIZE);
    s_Allocator.Initialize();

	NN_LOG("start\n");
    {
        imgdb::Initialize(s_Allocator);
        imgdb::search::Initialize(); // 写真列挙のための初期化

        TestSaveLoad();

        imgdb::search::Finalize();
        imgdb::Finalize();
    }
	NN_LOG("end\n");

    s_Allocator.Finalize();

	nn::cfg::Finalize();
}
